/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>

#include "screen_connector_launcher_service_evas/image_internal.h"
#include "screen_connector_launcher_service_evas/log_internal.h"

namespace screen_connector {

Image::Image(Evas_Object* raw) : EvasObject(raw, true) {
}

Image::~Image() = default;

void Image::Update(std::shared_ptr<RawImage> raw_image) {
  raw_image_ = std::move(raw_image);
  if (GetImageType() == ImageType::RemoteSurfaceTBM) {
    Update(raw_image_->GetWlBuffer());
  } else if (GetImageType() == ImageType::RemoteSurfaceImageFile) {
    Update(raw_image_->GetImageFile());
  } else if (GetImageType() == ImageType::SplashScreenImage) {
    Update(raw_image_->GetFilePath());
  } else if (GetImageType() == ImageType::SplashScreenEdje) {
    Update(raw_image_->GetFilePath(), raw_image_->GetFileGroup());
  }
}

void Image::Reset() {
  if (raw_image_.get() != nullptr &&
      raw_image_->GetImageType() == ImageType::RemoteSurfaceTBM) {
    evas_object_image_native_surface_set(GetRaw(), nullptr);
    raw_image_.reset();
  }
}

void Image::Update(std::shared_ptr<ImageFile> image_file) {
  void* map = mmap(NULL, image_file->GetSize(), PROT_READ, MAP_SHARED,
      image_file->GetRaw(), 0);

  evas_object_image_alpha_set(GetRaw(), EINA_TRUE);
  char format[] = "png";
  evas_object_image_memfile_set(GetRaw(), map, image_file->GetSize(),
      format, NULL);

  int width;
  int height;
  evas_object_image_size_get(GetRaw(), &width, &height);
  evas_object_image_fill_set(GetRaw(), 0, 0, width, height);
  evas_object_resize(GetRaw(), width, height);

  munmap(map, image_file->GetSize());
}

void Image::Update(std::shared_ptr<WlBuffer> wl_buffer) {
  tbm_surface_h tbm_surface = static_cast<tbm_surface_h>(
      wl_buffer_get_user_data(wl_buffer->GetRaw()));
  int width = tbm_surface_get_width(tbm_surface);
  int height = tbm_surface_get_height(tbm_surface);

  Evas_Native_Surface ns;
  memset(&ns, 0, sizeof(Evas_Native_Surface));
  ns.version = EVAS_NATIVE_SURFACE_VERSION;
  ns.type = EVAS_NATIVE_SURFACE_TBM;
  ns.data.tbm.buffer = tbm_surface;

  evas_object_image_alpha_set(GetRaw(), EINA_TRUE);
  evas_object_image_size_set(GetRaw(), width, height);
  evas_object_image_native_surface_set(GetRaw(), &ns);
  evas_object_image_pixels_dirty_set(GetRaw(), EINA_TRUE);
}

void Image::Update(std::string file_path) {
  evas_object_move(GetRaw(), 0, 0);
  evas_object_image_file_set(GetRaw(), file_path.c_str(), NULL);

  Evas_Load_Error err = evas_object_image_load_error_get(GetRaw());
  if (err != EVAS_LOAD_ERROR_NONE)
    _E("error(%d)", err);
}

void Image::Update(std::string file_path, std::string file_group) {
  evas_object_move(GetRaw(), 0, 0);
  edje_object_file_set(GetRaw(), file_path.c_str(), file_group.c_str());
}

void Image::SetSize(int32_t width, int32_t height) {
  evas_object_image_size_set(GetRaw(), width, height);
}

ImageType Image::GetImageType() const {
  return raw_image_->GetImageType();
}

Direction Image::GetDirection() const {
  return raw_image_->GetDirection();
}

int32_t Image::GetPositionX() const {
  return raw_image_->GetPositionX();
}

int32_t Image::GetPositionY() const {
  return raw_image_->GetPositionY();
}

std::string Image::GetSharedWidgetInfo() const {
  return raw_image_->GetSharedWidgetInfo();
}

}  // namespace screen_connector
