/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dlog.h>

#include "screen_connector_remote_surface/remote_surface_manager_internal.h"
#include "screen_connector_remote_surface/trs_mock_internal.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_REMOTE_SURFACE"

namespace screen_connector {

RemoteSurfaceManager& RemoteSurfaceManager::GetInst(bool mock) {
  static RemoteSurfaceManager sInst;

  if (!sInst.is_init_ && !mock)
    sInst.Init();

  return sInst;
}

ITRS* RemoteSurfaceManager::CreateTRS(int rid, std::shared_ptr<WlSurface> surf,
    bool mock) {
  struct tizen_remote_surface *tzrs = (surf != nullptr) ?
    tizen_remote_surface_manager_create_surface_with_wl_surface(rsm_,
        (uint32_t)rid, tbm_, surf->GetRaw()) :
    tizen_remote_surface_manager_create_surface(rsm_, (uint32_t)rid, tbm_);

  return mock ? new test::TRSMock() : new TRS(tzrs, true);
}

RemoteSurfaceManager::RemoteSurfaceManager() {
}

RemoteSurfaceManager::~RemoteSurfaceManager() {
  if (is_init_)
    Dispose();
}

void RemoteSurfaceManager::Dispose() {
  if (!is_init_)
    return;

  if (tbm_client_) {
    wayland_tbm_client_deinit(tbm_client_);
    tbm_client_ = nullptr;
  }

  if (rsm_) {
    tizen_remote_surface_manager_destroy(rsm_);
    rsm_ = nullptr;
  }

  LOGE("disconnect wl2_display");
  ecore_wl2_display_disconnect(ecore_wl2_connected_display_get(NULL));
  ecore_wl2_shutdown();
  is_init_ = false;
}

void RemoteSurfaceManager::Init() {
  struct wl_registry* registry;
  Ecore_Wl2_Global* global;
  Eina_Iterator *globals;

  if (is_init_)
    return;

  if (!ecore_wl2_init()) {
    LOGE("could not wl2 init");
    return;
  }

  ecore_wl2_display_connect(NULL);
  registry = ecore_wl2_display_registry_get(ecore_wl2_connected_display_get(NULL));
  globals = ecore_wl2_display_globals_get(ecore_wl2_connected_display_get(NULL));

  if (!registry || !globals) {
    LOGE("could not get registry(%p) or global list(%p)",
         registry, globals);

    LOGE("disconnect wl2_display");
    ecore_wl2_display_disconnect(ecore_wl2_connected_display_get(NULL));
    ecore_wl2_shutdown();
    if (globals)
      eina_iterator_free(globals);
    return;
  }

  EINA_ITERATOR_FOREACH(globals, global) {
    if (!strcmp(global->interface, "tizen_remote_surface_manager")) {
      rsm_ = (struct tizen_remote_surface_manager*)wl_registry_bind(registry,
             global->id, &tizen_remote_surface_manager_interface,
             global->version < 6 ? global->version : 6);
    }
  }

  if (!rsm_) {
    LOGE("could not get remote surface manager");
    LOGE("disconnect wl2_display");
    ecore_wl2_display_disconnect(ecore_wl2_connected_display_get(NULL));
    ecore_wl2_shutdown();
    eina_iterator_free(globals);
    return;
  }

  tbm_client_ = (struct wayland_tbm_client*)wayland_tbm_client_init(
                  ecore_wl2_display_get(ecore_wl2_connected_display_get(NULL)));
  if (!tbm_client_) {
    LOGE("could not get tbm client");
    tizen_remote_surface_manager_destroy(rsm_);
    rsm_ = nullptr;
    LOGE("disconnect wl2_display");
    ecore_wl2_display_disconnect(ecore_wl2_connected_display_get(NULL));
    ecore_wl2_shutdown();
    eina_iterator_free(globals);
    return;
  }

  tbm_ = (struct wl_tbm*)wayland_tbm_client_get_wl_tbm(tbm_client_);
  if (!tbm_) {
    LOGE("could not get tbm");
    wayland_tbm_client_deinit(tbm_client_);
    tbm_client_ = nullptr;
    tizen_remote_surface_manager_destroy(rsm_);
    rsm_ = nullptr;
    LOGE("disconnect wl2_display");
    ecore_wl2_display_disconnect(ecore_wl2_connected_display_get(NULL));
    ecore_wl2_shutdown();
    eina_iterator_free(globals);
    return;
  }
  eina_iterator_free(globals);
  is_init_ = true;
}

void RemoteSurfaceManager::Bind(const ITRS& trs, const WlSurface& suf) {
  if (trs.GetHandle() != nullptr) {
    tizen_remote_surface_manager_bind_surface(rsm_, suf.GetRaw(),
        trs.GetHandle());
  }
}

void RemoteSurfaceManager::Unbind(const ITRS& trs) {
  if (trs.GetHandle() != nullptr) {
    tizen_remote_surface_manager_bind_surface(rsm_, NULL, trs.GetHandle());
  }
}

}  // namespace screen_connector
