/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dlog.h>
#include <Ecore_Wl2.h>

#include <memory>

#include "screen_connector_remote_surface_evas/remote_surface_evas.h"
#include "screen_connector_remote_surface_evas/remote_surface_evas_implementation.h"
#include "screen_connector_remote_surface_evas/image_mock_internal.h"
#include "screen_connector_remote_surface_evas/util.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_REMOTE_SURFACE"
#define SC_TOOLKIT_HANDLE_TAG "SC_TOOLKIT_HANDLE_TAG"

namespace screen_connector {

RemoteSurfaceEvas::RemoteSurfaceEvas(const std::string& id,
                                     RemoteSurface::Type type,
                                     std::shared_ptr<EvasObject> viewerWin,
                                     bool mock)
    : RemoteSurface(id, type, util::GetWlSurface(*viewerWin.get()), mock),
      impl_(new Impl(this, viewerWin, mock)),
      win_ctx_(new WindowContext()) {
}

RemoteSurfaceEvas::RemoteSurfaceEvas(int rid, const std::string& id,
                                     RemoteSurface::Type type,
                                     std::shared_ptr<EvasObject> viewerWin,
                                     bool mock)
    : RemoteSurface(rid, id, type, util::GetWlSurface(*viewerWin.get()), mock),
      impl_(new Impl(this, viewerWin, mock)),
      win_ctx_(new WindowContext()) {
}

RemoteSurfaceEvas::~RemoteSurfaceEvas() = default;
RemoteSurfaceEvas::RemoteSurfaceEvas(RemoteSurfaceEvas&&) noexcept = default;
RemoteSurfaceEvas& RemoteSurfaceEvas::operator=(
    RemoteSurfaceEvas&&) noexcept = default;

RemoteSurfaceEvas::Impl::Impl(RemoteSurfaceEvas* parent,
    std::shared_ptr<EvasObject> viewerWin, bool mock)
    : viewer_win_(viewerWin), parent_(parent), mock_(mock) {
}

void RemoteSurfaceEvas::Impl::OnMouseIn(const EvasObject& obj,
                                        void* eventInfo) {
  Evas_Event_Mouse_In* ev =
      static_cast<Evas_Event_Mouse_In*>(eventInfo);
  const char* desc = evas_device_description_get(ev->dev);
  int x;
  int y;
  int w;
  int h;

  evas_object_geometry_get(obj.GetRaw(), &x, &y, &w, &h);
  if (desc == NULL)
    desc = "";

  LOGD("mouse in: %d %d", ev->canvas.x - x, ev->canvas.y - y);
  parent_->GetRaw()->TransferMouseEvent(
      TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_IN,
      0,
      0,
      ev->canvas.x - x,
      ev->canvas.y - y,
      0,
      0,
      0,
      0,
      evas_device_class_get(ev->dev),
      evas_device_subclass_get(ev->dev),
      desc,
      ev->timestamp);
}

void RemoteSurfaceEvas::Impl::OnMouseOut(const EvasObject& obj,
                                         void* eventInfo) {
  Evas_Event_Mouse_Out* ev =
      static_cast<Evas_Event_Mouse_Out*>(eventInfo);
  const char* desc = evas_device_description_get(ev->dev);
  int x;
  int y;
  int w;
  int h;

  evas_object_geometry_get(obj.GetRaw(), &x, &y, &w, &h);
  if (desc == NULL)
    desc = "";

  LOGD("mouse out: %d %d", ev->canvas.x - x, ev->canvas.y - y);
  parent_->GetRaw()->TransferMouseEvent(
      TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_OUT,
      0,
      0,
      ev->canvas.x - x,
      ev->canvas.y - y,
      0,
      0,
      0,
      0,
      evas_device_class_get(ev->dev),
      evas_device_subclass_get(ev->dev),
      desc,
      ev->timestamp);
}

void RemoteSurfaceEvas::Impl::OnMouseDown(const EvasObject& obj,
                                          void* eventInfo) {
  Evas_Event_Mouse_Down* ev =
      static_cast<Evas_Event_Mouse_Down*>(eventInfo);
  const char* desc = evas_device_description_get(ev->dev);
  int x;
  int y;
  int w;
  int h;

  if (parent_->IsDisabledInput(RemoteSurface::MOUSE_UP_DOWN)) {
    LOGD("mouse down disabled");
    return;
  }

  evas_object_geometry_get(obj.GetRaw(), &x, &y, &w, &h);

  LOGD("mouse down: %d %d", ev->canvas.x - x, ev->canvas.y - y);
  flick_ = false;

  if (parent_->GetType() == RemoteSurface::WATCH)
    prev_y_ = ev->canvas.y - y;

  if (desc == NULL)
    desc = "";

  parent_->GetRaw()->TransferMouseEvent(
      TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_DOWN,
      0,
      ev->button,
      ev->canvas.x - x,
      ev->canvas.y - y,
      wl_fixed_from_double(ev->radius_x),
      wl_fixed_from_double(ev->radius_y),
      wl_fixed_from_double(ev->pressure),
      wl_fixed_from_double(ev->angle),
      evas_device_class_get(ev->dev),
      evas_device_subclass_get(ev->dev),
      desc,
      ev->timestamp);
}

void RemoteSurfaceEvas::Impl::OnMouseUp(const EvasObject& obj,
                                        void* eventInfo) {
  Evas_Event_Mouse_Up* ev =
      static_cast<Evas_Event_Mouse_Up*>(eventInfo);
  const char *desc = evas_device_description_get(ev->dev);
  int x;
  int y;
  int w;
  int h;

  if (parent_->IsDisabledInput(RemoteSurface::MOUSE_UP_DOWN)) {
    LOGD("mouse up disabled");
    return;
  }

  if (ev->event_flags & EVAS_EVENT_FLAG_ON_HOLD) {
    LOGD("Event canceled");
    parent_->SendTouchCancel();
  }

  evas_object_geometry_get(obj.GetRaw(), &x, &y, &w, &h);
  LOGD("mouse up: %d %d", ev->canvas.x - x, ev->canvas.y - y);

  if (desc == NULL)
    desc = "";

  parent_->GetRaw()->TransferMouseEvent(
      TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_MOVE,
      0,
      0,
      ev->canvas.x - x,
      ev->canvas.y - y,
      wl_fixed_from_double(ev->radius_x),
      wl_fixed_from_double(ev->radius_y),
      wl_fixed_from_double(ev->pressure),
      wl_fixed_from_double(ev->angle),
      evas_device_class_get(ev->dev),
      evas_device_subclass_get(ev->dev),
      desc,
      ev->timestamp);

  parent_->GetRaw()->TransferMouseEvent(
      TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_UP,
      0,
      ev->button,
      ev->canvas.x - x,
      ev->canvas.y - y,
      wl_fixed_from_double(ev->radius_x),
      wl_fixed_from_double(ev->radius_y),
      wl_fixed_from_double(ev->pressure),
      wl_fixed_from_double(ev->angle),
      evas_device_class_get(ev->dev),
      evas_device_subclass_get(ev->dev),
      desc,
      ev->timestamp);
}

void RemoteSurfaceEvas::Impl::OnTouchMultiDown(const EvasObject& obj,
                                               void* eventInfo) {
  Evas_Event_Multi_Down* ev =
      static_cast<Evas_Event_Multi_Down*>(eventInfo);
  const char *desc = evas_device_description_get(ev->dev);
  int x;
  int y;
  int w;
  int h;

  if (parent_->IsDisabledInput(RemoteSurface::TOUCH_UP_DOWN)) {
    LOGD("touch-multi-down disabled");
    return;
  }

  evas_object_geometry_get(obj.GetRaw(), &x, &y, &w, &h);
  LOGD("multi down: %d %d", ev->canvas.x - x, ev->canvas.y - y);

  if (desc == NULL)
    desc = "";

  parent_->GetRaw()->TransferTouchEvent(
      TIZEN_REMOTE_SURFACE_EVENT_TYPE_TOUCH_DOWN,
      ev->device,
      1,
      ev->canvas.x - x,
      ev->canvas.y - y,
      wl_fixed_from_double(ev->radius_x),
      wl_fixed_from_double(ev->radius_y),
      wl_fixed_from_double(ev->pressure),
      wl_fixed_from_double(ev->angle),
      evas_device_class_get(ev->dev),
      evas_device_subclass_get(ev->dev),
      desc,
      ev->timestamp);
}

void RemoteSurfaceEvas::Impl::OnTouchMultiUp(const EvasObject& obj,
                                             void* eventInfo) {
  Evas_Event_Multi_Up* ev =
      static_cast<Evas_Event_Multi_Up*>(eventInfo);
  const char *desc = evas_device_description_get(ev->dev);
  int x;
  int y;
  int w;
  int h;

  if (parent_->IsDisabledInput(RemoteSurface::TOUCH_UP_DOWN)) {
    LOGD("touch-multi-up disabled");
    return;
  }

  evas_object_geometry_get(obj.GetRaw(), &x, &y, &w, &h);
  LOGD("multi up: %d %d", ev->canvas.x - x, ev->canvas.y - y);

  if (desc == NULL)
    desc = "";

  parent_->GetRaw()->TransferTouchEvent(
      TIZEN_REMOTE_SURFACE_EVENT_TYPE_TOUCH_UP,
      ev->device,
      1,
      ev->canvas.x - x,
      ev->canvas.y - y,
      wl_fixed_from_double(ev->radius_x),
      wl_fixed_from_double(ev->radius_y),
      wl_fixed_from_double(ev->pressure),
      wl_fixed_from_double(ev->angle),
      evas_device_class_get(ev->dev),
      evas_device_subclass_get(ev->dev),
      desc,
      ev->timestamp);
}

void RemoteSurfaceEvas::Impl::OnMouseMove(const EvasObject& obj,
                                          void* eventInfo) {
  Evas_Event_Mouse_Move* ev =
      static_cast<Evas_Event_Mouse_Move*>(eventInfo);
  const char *desc = evas_device_description_get(ev->dev);
  int x;
  int y;
  int w;
  int h;

  if (parent_->IsDisabledInput(RemoteSurface::MOUSE_MOVE_X_Y)) {
    LOGD("mouse-move disabled");
    return;
  }

  evas_object_geometry_get(obj.GetRaw(), &x, &y, &w, &h);
  LOGD("mouse move: %d %d", ev->cur.canvas.x - x, ev->cur.canvas.y - y);

  if (desc == NULL)
    desc = "";

  if (parent_->GetType() == RemoteSurface::WATCH) {
    if (flick_) {
      LOGD("flick event");
      return;
    }

    int deltaY = ev->cur.canvas.y - y - prev_y_;
    if (deltaY < 0)
      deltaY *= -1;

    if (deltaY > 10) {
      flick_ = true;
      return;
    }

    prev_y_ = ev->cur.canvas.y - y;
  }

  parent_->GetRaw()->TransferMouseEvent(
      TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_MOVE,
      0,
      0,
      ev->cur.canvas.x - x,
      ev->cur.canvas.y - y,
      wl_fixed_from_double(ev->radius_x),
      wl_fixed_from_double(ev->radius_y),
      wl_fixed_from_double(ev->pressure),
      wl_fixed_from_double(ev->angle),
      evas_device_class_get(ev->dev),
      evas_device_subclass_get(ev->dev),
      desc,
      ev->timestamp);
}

void RemoteSurfaceEvas::Impl::OnMouseWheel(const EvasObject& obj,
                                           void* eventInfo) {
  Evas_Event_Mouse_Wheel* ev =
      static_cast<Evas_Event_Mouse_Wheel*>(eventInfo);
  const char *desc = evas_device_description_get(ev->dev);

  if (parent_->IsDisabledInput(RemoteSurface::MOUSE_WHEEL)) {
    LOGD("mouse-wheel disabled");
    return;
  }

  LOGD("mouse wheel");
  if (desc == NULL)
    desc = "";

  parent_->GetRaw()->TransferMouseWheel(
      ev->direction,
      ev->z,
      evas_device_class_get(ev->dev),
      evas_device_subclass_get(ev->dev),
      desc,
      ev->timestamp);
}

void RemoteSurfaceEvas::Impl::OnShow(const EvasObject& obj, void* eventInfo) {
  LOGD("show");
  OnMove(obj, eventInfo);
}

void RemoteSurfaceEvas::Impl::OnHide(const EvasObject& obj, void* eventInfo) {
  LOGD("hide");
}

void RemoteSurfaceEvas::Impl::OnResize(const EvasObject& obj, void* eventInfo) {
  LOGD("resize");
  OnMove(obj, eventInfo);
}

void RemoteSurfaceEvas::Impl::OnMove(const EvasObject& obj, void *eventInfo) {
  if (!is_init_) {
    is_init_ = true;
    return;
  }

  Ecore_Wl2_Window *window = NULL;
  Ecore_Evas *ee;
  ee = ecore_evas_ecore_evas_get(
    evas_object_evas_get(viewer_win_->GetRaw()));
  window = ecore_evas_wayland2_window_get(ee);

  unsigned int win_id = ecore_wl2_window_id_get(window);
  if (parent_->IsVisibleArea() && !parent_->win_ctx_->IsFullyObscured(win_id)) {
    parent_->SendVisibility(true);
    visibility_ = RemoteSurfaceEvas::UNOBSCURED;
  } else {
    parent_->SendVisibility(false);
    visibility_ = RemoteSurfaceEvas::FULLY_OBSCURED;
  }
}

bool RemoteSurfaceEvas::IsVisibleArea() const {
  int x, y, w, h;
  Ecore_Wl2_Window *window = NULL;
  Ecore_Evas *ee;
  int window_x, window_y, window_w, window_h;
  int rotation;

  if (impl_->img_tbm_.get() == nullptr) {
    LOGE("img_tbm is NULL");
    return false;
  }

  ee = ecore_evas_ecore_evas_get(
    evas_object_evas_get(impl_->viewer_win_->GetRaw()));
  window = ecore_evas_wayland2_window_get(ee);
  ecore_wl2_window_geometry_get(window, &window_x, &window_y,
      &window_w, &window_h);
  evas_object_geometry_get(impl_->img_tbm_->GetRaw(), &x, &y, &w, &h);
  rotation = ecore_wl2_window_rotation_get(window);

  if (x >= 0 && x < window_w && y >= 0 && y < window_h &&
      (rotation == 0 || rotation == 180)) {
    LOGD("x %d, y %d w %d h %d, window_w %d window_h %d rotation %d",
        x, y, w, h, window_w, window_h, rotation);
    return true;
  } else if (x >= 0 && x < window_h && y >= 0 && y < window_w &&
      (rotation == 90 || rotation == 270)) {
    LOGD("x %d, y %d w %d h %d, window_w %d window_h %d rotation %d",
        x, y, w, h, window_w, window_h, rotation);
    return true;
  }

  return false;
}

void RemoteSurfaceEvas::OnBufferAdded(const std::string& appId,
                                      const std::string& instId, int pid) {
}

void RemoteSurfaceEvas::OnBufferRemoved(const std::string& appId,
                                        const std::string& instId, int pid) {
  if (impl_->img_tbm_.get() == nullptr)
    return;

  OnEvasRemoved(appId, instId, pid, *(impl_->img_tbm_));
  impl_->img_tbm_.reset();
  impl_->winVisibility_map_.clear();
}

void RemoteSurfaceEvas::OnBind() {
  impl_->img_tbm_->ClearTbm();
}

void RemoteSurfaceEvas::OnBufferChanged(int type, std::shared_ptr<WlBuffer> tbm,
                                        int fd, uint32_t size, uint32_t time) {
  bool isAdded = false;

  if (type != TIZEN_REMOTE_SURFACE_BUFFER_TYPE_TBM)
    return;

  if (tbm.get() == nullptr) {
      LOGW("Null tbm");
      return;
  }

  if (GetType() == RemoteSurface::WATCH && RemoteSurface::IsBound())
    return;

  if (impl_->img_tbm_.get() == nullptr) {
    LOGD("first added !!!! %d %s", type, GetAppId().c_str());
    impl_->winVisibility_map_.clear();
    impl_->img_tbm_.reset(
        impl_->mock_ ?
        new test::ImageMock(evas_object_image_filled_add(
            evas_object_evas_get(impl_->viewer_win_->GetRaw())),
            impl_.get(), GetInstId(), GetPid(), this)
        : new Image(evas_object_image_filled_add(
            evas_object_evas_get(impl_->viewer_win_->GetRaw())),
            impl_.get(), GetInstId(), GetPid(), this));
    Ecore_Evas *ee = ecore_evas_ecore_evas_get(
      evas_object_evas_get(impl_->viewer_win_->GetRaw()));
    Ecore_Wl2_Window* wlWin = ecore_evas_wayland2_window_get(ee);
    if (!wlWin) {
      LOGE("Failed to get wlWin");
    } else {
      unsigned int win_id = ecore_wl2_window_id_get(wlWin);
      impl_->winVisibility_map_[win_id] = GetVisibility(win_id);
    }
    isAdded = true;
  }

  impl_->img_tbm_->Update(tbm);
  if (isAdded)
    OnEvasAdded(GetAppId(), GetInstId(), GetPid(), *(impl_->img_tbm_));
  else
    OnEvasChanged(GetAppId(), GetInstId(), GetPid(), *(impl_->img_tbm_));
}

void RemoteSurfaceEvas::OnEvasAdded(const std::string& appId,
                                    const std::string& instId, int pid,
                                    const EvasObject& image) {
}

void RemoteSurfaceEvas::OnEvasRemoved(const std::string& appId,
                                      const std::string& instId, int pid,
                                      const EvasObject& image) {
}

void RemoteSurfaceEvas::OnEvasChanged(const std::string& appId,
                                      const std::string& instId, int pid,
                                      const EvasObject& image) {
}

void RemoteSurfaceEvas::Bind(const EvasObject& win) {
  struct wl_surface* surface;
  Ecore_Wl2_Window* wl_win;
  Ecore_Evas *ee = ecore_evas_ecore_evas_get(
    evas_object_evas_get(win.GetRaw()));
  int bind_win_id;

  wl_win = ecore_evas_wayland2_window_get(ee);
  if (!wl_win) {
    LOGE("failed to get wl_win");
    return;
  }

  surface = ecore_wl2_window_surface_get(wl_win);
  if (!surface) {
    LOGE("failed to get surface");
    return;
  }

  bind_win_id = ecore_wl2_window_id_get(wl_win);
  impl_->winVisibility_map_[bind_win_id] = GetVisibility(bind_win_id);
  impl_->bind_win_id_ = bind_win_id;

  RemoteSurface::Bind(std::shared_ptr<WlSurface>(
      new WlSurface(surface, false)));
}

void RemoteSurfaceEvas::Unbind() {
  impl_->winVisibility_map_.erase(impl_->bind_win_id_);
  RemoteSurface::Unbind();
}

void RemoteSurfaceEvas::SetChangedEventFilter(
    RemoteSurface::ChangedEventFilter filter) {
  RemoteSurface::SetChangedEventFilter(filter);
  if (filter != FILTER_ALL && filter != FILTER_TBM) {
    RemoteSurface::SetBlock(false);
  } else {
    RemoteSurface::SetBlock(true);
    RemoteSurface::SendTouchCancel();
  }
}

void RemoteSurfaceEvas::SendMouseUp() {
  uint32_t timestamp;
  int x;
  int y;
  int w;
  int h;

  if (impl_->img_tbm_.get() == nullptr)
    return;

  evas_object_geometry_get(impl_->img_tbm_->GetRaw(), &x, &y, &w, &h);
  timestamp = (uint32_t)((uint64_t)(ecore_time_get() * 1000.0) & 0xffffffff);
  RemoteSurface::SendMouseUp(w / 2, h / 2, timestamp);
}

void RemoteSurfaceEvas::NotifyWindowVisibilityEvent(int type, void* event) {
  if (IsFrozen())
    return;

  if (!IsVisibleArea())
    return;

  Ecore_Wl2_Event_Window_Visibility_Change* ev =
      static_cast<Ecore_Wl2_Event_Window_Visibility_Change*>(event);
  Visibility v;

  if (ev->fully_obscured)
    v = FULLY_OBSCURED;
  else
    v = UNOBSCURED;

  impl_->winVisibility_map_[ev->win] = v;
  v = FULLY_OBSCURED;

  for (auto& p : impl_->winVisibility_map_) {
    if (p.second == UNKNOWN) {
      LOGE("viewer not ready");
      return;
    } else if (p.second == UNOBSCURED) {
      v = UNOBSCURED;
    }
  }

  if (v == FULLY_OBSCURED)
    SendVisibility(false);
  else
    SendVisibility(true);

  impl_->visibility_ = v;
}

RemoteSurfaceEvas::Visibility RemoteSurfaceEvas::GetVisibility() const {
  return impl_->visibility_;
}

RemoteSurfaceEvas::Visibility RemoteSurfaceEvas::GetVisibility(
    unsigned int win) {
  if (!win_ctx_->FindWindow(win)) {
    LOGE("Failed to find window(%u)", win);
    return RemoteSurfaceEvas::UNKNOWN;
  }

  if (win_ctx_->IsFullyObscured(win))
    return RemoteSurfaceEvas::FULLY_OBSCURED;

  return RemoteSurfaceEvas::UNOBSCURED;
}

}  // namespace screen_connector
