/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <dlog.h>
#include <wayland-client.h>
#include <Elementary.h>
#include <Ecore_Wl2.h>
#include <wayland-extension/tizen-extension-client-protocol.h>
#include <tizen-remote-surface-client-protocol.h>

#include <map>
#include <list>
#include <algorithm>
#include <memory>

#include "screen_connector_toolkit.h"
#include "screen_connector_remote_surface/remote_surface.h"
#include "screen_connector_remote_surface/trs_interface.h"
#include "screen_connector_remote_surface/handle.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_TOOLKIT"

class RemoteSurfaceToolkit : public screen_connector::RemoteSurface {
 public:
  RemoteSurfaceToolkit(const std::string& id,
                       screen_connector::RemoteSurface::Type type,
                       screen_connector_toolkit_ops *ops,
                       std::shared_ptr<screen_connector::WlSurface> wl_surface,
                       void* data)
      : screen_connector::RemoteSurface(id, type, wl_surface),
        ops_(*ops), data_(data) {
  }


  RemoteSurfaceToolkit(const std::string& id,
                       screen_connector::RemoteSurface::Type type,
                       screen_connector_toolkit_ops *ops, int rid,
                       std::shared_ptr<screen_connector::WlSurface> wl_surface,
                       void* data)
      : screen_connector::RemoteSurface(rid, id, type, wl_surface),
        ops_(*ops), data_(data) {
  }

  void* GetUserData() {
    return data_;
  }

  void Unbind() override {
    screen_connector::RemoteSurface::Unbind();
    screen_connector::RemoteSurface::GetRaw()->Redirect();
  }

  void OnBufferAdded(const std::string& appId, const std::string& instId,
                     int pid) override {
    if (ops_.added_cb)
      ops_.added_cb(appId.c_str(), instId.c_str(), pid, data_);
  }

  void OnBufferRemoved(const std::string& appId, const std::string& instId,
                       int pid) override {
    if (ops_.removed_cb)
      ops_.removed_cb(appId.c_str(), instId.c_str(), pid, data_);
  }

  void OnBufferChanged(int type,
                       std::shared_ptr<screen_connector::WlBuffer> tbm, int fd,
                       uint32_t size, uint32_t time) override {
    if (ops_.updated_cb) {
      ops_.updated_cb(screen_connector::RemoteSurface::GetRaw()->GetHandle(),
          type, tbm->GetRaw(), fd, size, time, nullptr,
          screen_connector::RemoteSurface::GetAppId().c_str(),
          screen_connector::RemoteSurface::GetInstId().c_str(),
          screen_connector::RemoteSurface::GetPid(), data_);
      prev_tbms_.push_back(std::move(tbm));
    }
  }

  void DisposeBuffer(const struct wl_buffer* tbm) {
    for (auto& i : prev_tbms_) {
      if (i->GetRaw() == tbm) {
        prev_tbms_.remove(i);
        return;
      }
    }
  }

 private:
  screen_connector_toolkit_ops ops_;
  void* data_;
  std::list<std::shared_ptr<screen_connector::WlBuffer>> prev_tbms_;
};

class Toolkit {
 public:
  Toolkit() { }
  ~Toolkit() { }

  void Add(RemoteSurfaceToolkit* rs) {
    toolkits_.emplace_back(rs);
  }

  void Remove(RemoteSurfaceToolkit* rs) {
    for (auto& i : toolkits_) {
      if (i.get() == rs) {
        toolkits_.remove(i);
        break;
      }
    }
  }

  bool Exist(const std::string& id) {
    for (auto& i : toolkits_) {
      if (i.get()->GetInstId() == id)
        return true;
    }
     return false;
  }

 private:
  std::list<std::unique_ptr<RemoteSurfaceToolkit>> toolkits_;
};

static std::map<int, Toolkit*> __toolkits;

extern "C" EXPORT_API int
screen_connector_toolkit_init(screen_connector_screen_type_e type) {
  if (__toolkits.find(type) != __toolkits.end())
    return 0;

  screen_connector::RemoteSurface::InitManager();
  __toolkits[static_cast<int>(type)] = new Toolkit();

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_fini(screen_connector_screen_type_e type) {
  auto i = __toolkits.find(type);

  if (i == __toolkits.end())
    return 0;

  __toolkits.erase(i);

  if (__toolkits.empty())
    screen_connector::RemoteSurface::FiniManager();

  return 0;
}

extern "C" EXPORT_API screen_connector_toolkit_h
screen_connector_toolkit_add_by_rid(screen_connector_toolkit_ops *ops,
                                    const char* id,
                                    screen_connector_screen_type_e type,
                                    int surface_id,
                                    struct wl_surface* surface,
                                    void* data) {
  if (screen_connector_toolkit_is_exist(id, type)) {
    LOGE("Already exists %s", id);
    return NULL;
  }

  auto rs = new RemoteSurfaceToolkit(id,
      (screen_connector::RemoteSurface::Type)type, ops, surface_id,
      std::shared_ptr<screen_connector::WlSurface>(
        new screen_connector::WlSurface(surface, false)), data);

  __toolkits[static_cast<int>(type)]->Add(rs);

  return static_cast<void*>(rs);
}

extern "C" EXPORT_API screen_connector_toolkit_h
screen_connector_toolkit_add(screen_connector_toolkit_ops *ops,
                             const char* id,
                             screen_connector_screen_type_e type,
                             void* data) {
  if (screen_connector_toolkit_is_exist(id, type)) {
    LOGE("Already exists %s", id);
    return NULL;
  }

  auto rs = new RemoteSurfaceToolkit(id,
      (screen_connector::RemoteSurface::Type)type, ops, nullptr, data);

  __toolkits[static_cast<int>(type)]->Add(rs);

  return static_cast<void*>(rs);
}

extern "C" EXPORT_API screen_connector_toolkit_h
screen_connector_toolkit_add_with_surface(screen_connector_toolkit_ops *ops,
                             const char* id,
                             screen_connector_screen_type_e type,
                             struct wl_surface* surface,
                             void* data) {
  if (screen_connector_toolkit_is_exist(id, type)) {
    LOGE("Already exists %s", id);
    return NULL;
  }

  auto rs = new RemoteSurfaceToolkit(id,
      (screen_connector::RemoteSurface::Type)type,
      ops, std::shared_ptr<screen_connector::WlSurface>(
        new screen_connector::WlSurface(surface, false)), data);

  __toolkits[static_cast<int>(type)]->Add(rs);

  return static_cast<void*>(rs);
}

extern "C" EXPORT_API bool
screen_connector_toolkit_is_exist(const char* id,
                                  screen_connector_screen_type_e type) {
  auto i = __toolkits.find(type);

  if (i == __toolkits.end())
    return false;

  if (__toolkits[static_cast<int>(type)]->Exist(id))
    return true;

  return false;
}

extern "C" EXPORT_API int
screen_connector_toolkit_remove(screen_connector_toolkit_h handle) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);

  for (int i = SCREEN_CONNECTOR_SCREEN_TYPE_WIDGET;
      i <= SCREEN_CONNECTOR_SCREEN_TYPE_ALL; i++) {
    auto sci = __toolkits.find(i);

    if (sci == __toolkits.end())
      continue;

    __toolkits[i]->Remove(rs);
  }

  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_update(const char* appid, const char* instance_id) {
  return aul_screen_connector_send_update_request(appid, instance_id);
}

extern "C" EXPORT_API const char*
screen_connector_toolkit_get_appid(screen_connector_toolkit_h handle) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  return rs->GetAppId().c_str();
}

extern "C" EXPORT_API const char*
screen_connector_toolkit_get_instance_id(screen_connector_toolkit_h handle) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  return rs->GetInstId().c_str();
}

extern "C" EXPORT_API int
screen_connector_toolkit_get_pid(screen_connector_toolkit_h handle) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  return rs->GetPid();
}

extern "C" EXPORT_API int
screen_connector_toolkit_get_surface_id(screen_connector_toolkit_h handle) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  return rs->GetRid();
}

extern "C" EXPORT_API void*
screen_connector_toolkit_get_user_data(screen_connector_toolkit_h handle) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  return rs->GetUserData();
}

extern "C" EXPORT_API unsigned int
screen_connector_toolkit_get_event_filter(screen_connector_toolkit_h handle) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  return (unsigned int)(rs->GetChangedEventFilter());
}

extern "C" EXPORT_API int
screen_connector_toolkit_check_input_disabled(screen_connector_toolkit_h handle,
    screen_connector_input_type_e input) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  auto f = rs->GetChangedEventFilter();
  return ((((screen_connector_input_type_e)f) & input) > 0);
}

extern "C" EXPORT_API struct tizen_remote_surface*
screen_connector_toolkit_get_trs(screen_connector_toolkit_h handle) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  return rs->GetRaw()->GetHandle();
}

extern "C" EXPORT_API int
screen_connector_toolkit_bind(screen_connector_toolkit_h handle,
                              struct wl_surface* surface) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  rs->Bind(std::shared_ptr<screen_connector::WlSurface>(
      new screen_connector::WlSurface(surface, false)));
  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_unbind(screen_connector_toolkit_h handle) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  rs->Unbind();
  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_set_changed_event_filter(
    screen_connector_toolkit_h handle,
    screen_connector_changed_event_filter_type type) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  rs->SetChangedEventFilter(
      (screen_connector::RemoteSurface::ChangedEventFilter)type);
  return 0;
}

extern "C" EXPORT_API int
screen_connector_toolkit_dispose_buffer(screen_connector_toolkit_h handle,
    const struct wl_buffer* tbm) {
  RemoteSurfaceToolkit* rs = static_cast<RemoteSurfaceToolkit*>(handle);
  rs->DisposeBuffer(tbm);
  return 0;
}
